// osimHybrid.cpp
// Author: Tom Erez
/*
 * Copyright (c) 2005, Stanford University. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * - Neither the name of the Stanford University nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <OpenSim/OpenSim.h>
#include <OpenSim/Common/IO.h>
#include "SimTKmath.h"
#include "mex.h"
#include "SimTKsimbody.h"

using namespace OpenSim;
using namespace std;
//using namespace SimTK;

class VectorConstantController : public Controller {
public:
   double *controlVector;

   double computeControl( const SimTK::State& s, int index ) const{
      return controlVector[index];
   }

   void setControlVector( double *cvec ) {
      for (int i=0;i<_numControls;++i) controlVector[i]=cvec[i];
   }
	/**
	 * Constructor
	 *
	 * @param aModel Model to be controlled
	 */
	VectorConstantController( Model& aModel ) :
	  Controller( aModel ) {}

	/**
	 * This function connects the model actuators to this controller.
	 *
	 * @param as Actuator set to be controlled
	 */
	void setActuators( Set<Actuator>& as ) {
		int i, j;
		for( i = 0, j = 0; i < as.getSize(); i++ ) {
			Actuator& act = as.get(i);
			if( act.getControlIndex() == -1 ) { // check if already controlled
				act.setController( this );
				act.setIsControlled( true );
				act.setControlIndex( j++ );
			}
		}
      
      
      //what about updating _numControls?
      
      
      if (controlVector!=NULL) delete controlVector;
      controlVector = new double[as.getSize()];
	}

	/**
	 * A function called during system setup
	 *
	 * @param aModel Model to be controlled
	 */
	void setup( Model& aModel )  {
		 _model = &aModel;
	}

	/**
	 * Another function called during system setup
	 *
	 * @param aModel Model to be controlled
	 */
	void setupSystem( SimTK::MultibodySystem& system ) { 
		_actuatorSet.setSize(0);
      /*
		cout << "\n CustomController::setupSystem begin num Forces=" <<  _model->getForceSet().getSize() << endl;
		for(int i=0; i< _model->getForceSet().getSize(); i++) {
			cout << "Force " << i << " = "<< _model->getForceSet().get(i).getName() << endl;
		}
		cout << "\n CustomController::setupSystem begin num Actuators=" <<  _model->getActuators().getSize() << endl;
		for(int i=0; i< _model->getActuators().getSize(); i++) {
			cout << "Actuator " << i << " = "<< _model->getActuators().get(i).getName() << endl;
		}
		cout << endl;
       */
		// Add these actuators to the model and set their indexes 
		const Set<Actuator>& cs = _model->getActuators();
		for(int i=0; i<cs.getSize(); i++) {
			// cout << " CustomController::setupSystem " <<  cs.get(i).getName()+"_corrector" << "  added " << endl;
			Actuator *actuator = &cs.get(i);
			actuator->setController(this);
			actuator->setIsControlled(true);
			actuator->setControlIndex(i);
			_actuatorSet.append(actuator);
		}
		_numControls = _actuatorSet.getSize();
      if (controlVector!=NULL) delete controlVector;
      controlVector = new double[_numControls];
		// printf(" CustomController::setupSystem end  num Actuators= %d \n",  _model->getForceSet().getSize() );
	}
};


int *nstates;
int *ncontrols;
int nmarkers;
int *nU;
int nmodelsCur=0;
double dt;
Model* *model;
SimTK::Integrator* *integrator;
SimTK::TimeStepper* *timestepper;
SimTK::State* state;
VectorConstantController* *controller;


int setup(double dtn, double acc, string modelFile[],int nmodels)
{
   try {
      
      Thelen2003Muscle *crap = new Thelen2003Muscle();
      delete crap;
            
      string saveWorkingDirectory = IO::getCwd();
      string directoryOfSetupFile = IO::getParentDirectory(modelFile[0]);
      IO::chDir(directoryOfSetupFile);
/*      cout << nmodelsCur << endl;
      if (nmodelsCur!=0)
      {
         cout << "!" << endl;
         delete[] nstates;
         delete[] ncontrols;
         delete[] nU;
         for (int i=0;i<nmodelsCur;++i)
         {
            cout << i << "@:2" << endl;
            delete controller[i];
            cout << i << "@:5" << endl;
            delete model[i];
            cout << i << "@:6" << endl;
            delete timestepper[i];
            cout << i << "@:3" << endl;
            delete integrator[i];
            cout << i << "@:4" << endl;
         }
         delete[] model;
         delete[] integrator;
         delete[] timestepper;
         delete[] state;
         delete[] controller;
      }
*/      
      nmodelsCur=nmodels;
      nstates = new int[nmodels];
      ncontrols = new int[nmodels];
      nU = new int[nmodels];
      model = new Model*[nmodels];
      integrator = new SimTK::Integrator*[nmodels];
      timestepper = new SimTK::TimeStepper*[nmodels];
      state = new SimTK::State[nmodels];
      controller = new VectorConstantController*[nmodels];
      dt=dtn;
      for (int i=0;i<nmodels;++i)
      {
         model[i] = new Model(modelFile[i]);
         controller[i] = new VectorConstantController(*model[i]);
         controller[i]->setupSystem(model[i]->getSystem());
         model[i]->addController( controller[i] );
         state[i] = model[i]->initSystem();

         
         model[i]->printDetailedInfo(state[i],cout);

         nstates[i] = model[i]->getNumStates();
         ncontrols[i] = model[i]->getActuators().getSize();
         nmarkers = model[i]->getMarkerSet().getSize();
         nU[i] = state[i].getNU();
         integrator[i] = new SimTK::RungeKuttaFeldbergIntegrator(model[i]->getSystem());
         integrator[i]->setAccuracy(acc);
         timestepper[i] = new SimTK::TimeStepper(model[i]->getSystem(), *integrator[i]);
      }
      IO::chDir(saveWorkingDirectory);
      nmodelsCur=nmodels;
   } catch(Exception x) {
      x.print(cout);
      return(-1);
   }
   
   return(0);
}

 void getGRF(int modelInx, double grf[])
{
	try {
		model[modelInx]->getSystem().realize(state[modelInx],SimTK::Stage::Dynamics);
	} catch(exception &x) {
		cout<<x.what()<<endl;
		cout<<"realize dynamics error"<<endl;
		return;
	}
   int nb = model[modelInx]->getNumBodies();
   SimTK::Vector_<SimTK::Vec3> allForcesVec(nb);
   SimTK::Vector_<SimTK::Vec3> allMomentsVec(nb);
   model[modelInx]->getSimbodyEngine().computeReactions(state[modelInx], allForcesVec, allMomentsVec);
   SimTK::Vec3 rf=allForcesVec[0];
   SimTK::Vec3 rm=allMomentsVec[0];
   grf[0]=rm[0];
   grf[1]=rm[1];
   grf[2]=rm[2];
   grf[3]=rf[0];
   grf[4]=rf[1];
   grf[5]=rf[2];
//   cout << rm[0] << ";" << rm[1] << "h" << rm[2] << "a" << rf[1] << "%" << allForcesVec[0][0] << endl;
}
 
 void getmarkers(int modelInx, double markerposs[], double markerv[])
{
   MarkerSet markerset = model[modelInx]->getMarkerSet();
   BodySet bodyset = model[modelInx]->getBodySet();
   Body b;
   Marker m;
   for (int i=0;i<nmarkers;i++)
   {
      m=markerset.get(i);
      b = bodyset.get(*(m.getBodyName()));
      SimTK::Vec3 res;
      model[modelInx]->getSimbodyEngine().getPosition(state[modelInx],b,m.getOffset(),res);
      for (int j=0;j<3;j++) markerposs[i*3+j]=res[j];
      if (markerv!=0)
      {
         model[modelInx]->getSimbodyEngine().getVelocity(state[modelInx],b,m.getOffset(),res);
         for (int j=0;j<3;j++) markerv[i*3+j]=res[j];
      }
   }
}



void setState(int modelInx, double stateVec[])
{
   SimTK::Vector Q(nU[modelInx]);
   SimTK::Vector U(nU[modelInx]);
   SimTK::Vector Z(nstates[modelInx]-2*nU[modelInx]);
   for (int i=0;i<nU[modelInx];++i)
   {
      Q[i]=stateVec[i];
      U[i]=stateVec[i+nU[modelInx]];
   }
   for (int i=0;i<nstates[modelInx]-2*nU[modelInx];++i)
   {
      Z[i]=stateVec[i+2*nU[modelInx]];
   }
   state[modelInx].setQ(Q);
   state[modelInx].setU(U);
   state[modelInx].setZ(Z);
   state[modelInx].setTime(0);
   model[modelInx]->getSystem().realize(state[modelInx],SimTK::Stage::Velocity);
}

void step(int modelInx,double newStateVec[],double dtin)
{
   timestepper[modelInx]->initialize(state[modelInx]);
   SimTK::Integrator::SuccessfulStepStatus status = timestepper[modelInx]->stepTo(dtin);
//   state[modelInx] = integrator[modelInx]->getState();
//    cout << integrator[modelInx]->getState().getSystemStage().getName() << endl;
   SimTK::Vector Y = integrator[modelInx]->getState().getY();
   for (int i=0;i<nstates[modelInx];++i) newStateVec[i]=Y[i];
}

void getDerivs(int modelInx,double dxdt[])
{
   try {
      model[modelInx]->getSystem().realize(state[modelInx],SimTK::Stage::Acceleration);
   } catch(exception &x) {
      cout<<x.what()<<endl;
      cout<<"SimbodyEngine.computeDerivatives: invalid derivatives."<<endl;
      return;
   }
   SimTK::Vector dY = state[modelInx].getYDot();
   for (int i=0;i<nstates[modelInx];++i) dxdt[i]=dY[i];
}

//new ground point:
//modify model properties
//re-initialize system
//OR
//combine a slider constraint and perscribed motion with velocity 0 and tunable position


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
   if (nlhs == 0)
	{
      if (nrhs<2) mexErrMsgTxt("if expecting num output, use lhs; if configuring a model, give dt");
      if (!mxIsDouble(prhs[0]) || !mxIsDouble(prhs[1]) || mxIsDouble(prhs[2])) mexErrMsgTxt("dt is first, then integrator accuracy, and finally as many filenames as you want (at least one!)");
      int mrows1 = mxGetM(prhs[0]);
      int ncols1 = mxGetN(prhs[0]);
      if (mrows1>1 || ncols1 >1)
      {
         mexErrMsgTxt("size/type error - dt");
      }
      double *dtp = mxGetPr(prhs[0]);
      double dtn = dtp[0];
      mrows1 = mxGetM(prhs[0]);
      ncols1 = mxGetN(prhs[0]);
      if (mrows1>1 || ncols1 >1)
      {
         mexErrMsgTxt("size/type error - accuracy");
      }
      double *accp = mxGetPr(prhs[1]);
      double acc = accp[0];
      string *str=new string[nrhs-2];
   	for (int i=2;i<nrhs;i++)
      {
         
         str[i-2] = mxArrayToString(prhs[i]);
      }
  		setup(dtn,acc,str,nrhs-2);
  		
  		
  		
   } else {
   
      if (nrhs>4) mexErrMsgTxt("four rhs and six lhs arguments at most");
      if (nrhs<2) mexErrMsgTxt("two rhs at least (index and states).  if you initiated only one model, its index is 1.");
      double *tModelIndex = mxGetPr(prhs[0]);
      int modelInx=  (int)tModelIndex[0]-1;
      if (modelInx>=nmodelsCur) mexErrMsgTxt("no such model");
      
      
      int mrows0 = mxGetM(prhs[1]);
      int ncols0 = mxGetN(prhs[1]);
      if (!(mxIsDouble(prhs[1]) && (mrows0 == nstates[modelInx])))
      {
         cout<<nstates[modelInx]<<endl;
         mexErrMsgTxt("size/type error - states");
      }
      double *stateVec = mxGetPr(prhs[1]);
      int markerdims[3];
      markerdims[0]=3;
      markerdims[1]=nmarkers;
      markerdims[2]=ncols0;
      double *markerposs=0;
      double *markerv=0;
      if (nrhs>2)
      {
         plhs[0] = mxCreateDoubleMatrix(nstates[modelInx], ncols0, mxREAL);
         double *dxdt = mxGetPr(plhs[0]);
         double *grf=0;
         double *xnew=0;
         if (nlhs>1)
         {
            plhs[1] = mxCreateNumericArray(3, markerdims, mxDOUBLE_CLASS, mxREAL);
            markerposs = mxGetPr(plhs[1]);
            if (nlhs>2)
            {
               plhs[2] = mxCreateNumericArray(3, markerdims, mxDOUBLE_CLASS, mxREAL);
               markerv = mxGetPr(plhs[2]);
               if (nlhs>3)
               {
                  int gdims[3];
                  gdims[0]=6;
                  gdims[1]=1;
                  gdims[2]=ncols0;
                  plhs[3] = mxCreateNumericArray(3, gdims, mxDOUBLE_CLASS, mxREAL);
                  grf = mxGetPr(plhs[3]);
                  if (nlhs>4)
                  {
                     plhs[4] = mxCreateDoubleMatrix(nstates[modelInx], ncols0, mxREAL);
                     xnew = mxGetPr(plhs[4]);
                  }
               }
            }
         }
         if (ncontrols==0) mexErrMsgTxt("cant handle models without controls");
         int mrows1 = mxGetM(prhs[2]);
         int ncols1 = mxGetN(prhs[2]);
         if (!(mxIsDouble(prhs[2]) && (mrows1 == ncontrols[modelInx])))
         {
            cout<<ncontrols[modelInx]<<endl;
            mexErrMsgTxt("size/type error - controls");
         }
         if (ncols0 != ncols1) mexErrMsgTxt("size error - inputs not correlated");
         double *control = mxGetPr(prhs[2]);      

         double dtnow;
         if (nrhs<4)
         {
            dtnow=dt;
         } else {
            int mrows4 = mxGetM(prhs[3]);
            if (!(mxIsDouble(prhs[3]) && (mrows4 == 1)))
            {
               mexErrMsgTxt("fourth input must be scalar (dt)");
            }
            double *dtnowin = mxGetPr(prhs[3]);
            dtnow = dtnowin[0];
         }
         
         for (int i=0;i<ncols0;++i)
         {
            setState(modelInx, &stateVec[i*nstates[modelInx]]);
            controller[modelInx]->setControlVector(&control[i*ncontrols[modelInx]]);
            getDerivs(modelInx,&dxdt[i*nstates[modelInx]]);
            //see what else is needed
            if (nlhs>1)
            {
               if (nlhs>2)
               {
                  getmarkers(modelInx,&markerposs[i*3*nmarkers],&markerv[i*3*nmarkers]);
               } else {
                  getmarkers(modelInx,&markerposs[i*3*nmarkers],markerv);
               }
               if (nlhs>3)
               {
                  getGRF(modelInx,&grf[i*6]);
                  if (nlhs>4) step(modelInx,&xnew[i*nstates[modelInx]],dtnow);
               }
            } //whatever else was needed
         } //end for
         
      } else { // no control
         plhs[0] = mxCreateNumericArray(3, markerdims, mxDOUBLE_CLASS, mxREAL);
         double *markerposs = mxGetPr(plhs[0]);      
         double *markerv=0;
         if (nlhs>1)
         {
            plhs[1] = mxCreateNumericArray(3, markerdims, mxDOUBLE_CLASS, mxREAL);
            markerv = mxGetPr(plhs[1]);
         }
         for (int i=0;i<ncols0;++i)
         {
            setState(modelInx,&stateVec[i*nstates[modelInx]]);
            if (nlhs>1)
            {
               getmarkers(modelInx,&markerposs[i*3*nmarkers],&markerv[i*3*nmarkers]);            
            } else {
               getmarkers(modelInx,&markerposs[i*3*nmarkers],markerv);
            }
         }
         if (nlhs>2) mexErrMsgTxt("you get the mass/GRF/integration only when you give actuation");
     }
   }
}
