// osimHybrid.cpp
// Author: Tom Erez
/*
 * Copyright (c) 2005, Stanford University. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * - Neither the name of the Stanford University nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <OpenSim/OpenSim.h>
#include <OpenSim/Common/IO.h>
#include "SimTKmath.h"
#include "mex.h"

using namespace OpenSim;
using namespace std;
//using namespace SimTK;

int nstates;
int ncontrols;
int nU;
Model *model;
SimTK::State state;

int setup(string modelFile)
{
   try {
      
	//LoadOpenSimLibrary("osimSimbodyEngine");
	LoadOpenSimLibrary("osimActuators");

      Thelen2003Muscle *crap = new Thelen2003Muscle();
      delete crap;
      string saveWorkingDirectory = IO::getCwd();
      string directoryOfSetupFile = IO::getParentDirectory(modelFile);
      IO::chDir(directoryOfSetupFile);
      
      model = new Model(modelFile);
      state = model->initSystem();

      model->printDetailedInfo(state,cout);
      nstates = model->getNumStates();
      ncontrols = model->getActuators().getSize();
      nU = state.getNU();
      IO::chDir(saveWorkingDirectory);
      
   } catch(Exception x) {
      x.print(cout);
      return(-1);
   }
   
   return(0);
}

void setState(double stateVec[])
{
   SimTK::Vector Q(nU);
   SimTK::Vector U(nU);
   SimTK::Vector Z(nstates-2*nU);
   for (int i=0;i<nU;++i)
   {
      Q[i]=stateVec[i];
      U[i]=stateVec[i+nU];
   }
   for (int i=0;i<nstates-2*nU;++i)
   {
      Z[i]=stateVec[i+2*nU];
   }
   state.setQ(Q);
   state.setU(U);
   state.setZ(Z);
   state.setTime(0);
   model->getSystem().realize(state,SimTK::Stage::Position);
}

void getLengths(double mslLengths[], double mArms[], mxArray * Paths)
{
   Muscle *mus;
   OpenSim::Array<PathPoint*> pathArray;
   SimTK::Vec3 globalLocation;
   PathPoint* nextPoint;
   OpenSim::Array<double> momentArms;
   for(int j=0;j<ncontrols;j++)  {
       mus = dynamic_cast<Muscle*>( &(model->getForceSet().get(j))); 
       if( mus ){
          mslLengths[j]=mus->getLength(state);
          if (mArms!=0)
          {
            for (int i=0; i<nU;++i)
            {
               //GeometryPath().computeMomentArms(state,momentArms);
               mArms[j+ncontrols*i]=mus->computeMomentArm(state,model->getCoordinateSet()[i]);
            }
          }
          pathArray = mus->getGeometryPath().getCurrentPath(state);
          mxArray *globalPositions = mxCreateNumericMatrix((mwSize)3,(mwSize)pathArray.getSize(),mxDOUBLE_CLASS, mxREAL);          
          mxArray *localPositions = mxCreateNumericMatrix((mwSize)3,(mwSize)pathArray.getSize(),mxDOUBLE_CLASS, mxREAL);
          mxArray *bodies = mxCreateNumericMatrix((mwSize)1,(mwSize)pathArray.getSize(),mxDOUBLE_CLASS, mxREAL);
          double *gp = mxGetPr(globalPositions);
          double *lp = mxGetPr(localPositions);
          double *b = mxGetPr(bodies);
          for (int i = 0; i < pathArray.getSize(); i++){
             nextPoint =pathArray.get(i);
             // xform point to global frame
             const SimTK::Vec3& location=nextPoint->getLocation();
             const OpenSim::Body& body = nextPoint->getBody();
             model->getSimbodyEngine().transformPosition(state, body, location, globalLocation);
             gp[i*3]=globalLocation[0];
             gp[i*3+1]=globalLocation[1];
             gp[i*3+2]=globalLocation[2];
             lp[i*3]=location[0];
             lp[i*3+1]=location[1];
             lp[i*3+2]=location[2];
             b[i]=body.getIndex();            
          }
          if (Paths)
          {
             mxSetCell(Paths,(mwIndex)j,globalPositions);
             mxSetCell(Paths,(mwIndex)(model->getForceSet().getSize()+j),localPositions);
             mxSetCell(Paths,(mwIndex)(2*(model->getForceSet().getSize())+j),bodies);
          }
      }
   }
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nlhs == 0)
	{
      if (mxIsDouble(prhs[0])) mexErrMsgTxt("if you're hoping for numerical output, give us some output variables");
      string str = mxArrayToString(prhs[0]);
  		setup(str);
   }
   else
   {
      if (nrhs>1) mexErrMsgTxt("only state as input");
      if (nlhs>3) mexErrMsgTxt("this only outputs muscle lengths, paths, and moment arms");
      // Check that the inputs are the correct format.
      int mrows = mxGetM(prhs[0]);
      int ncols = mxGetN(prhs[0]);
      if (!(mxIsDouble(prhs[0]) && (mrows == nstates)))
      {
         cout<<nstates<<endl;
         mexErrMsgTxt("size/type error - input1");
      }
      double *stateVec = mxGetPr(prhs[0]);
      
      mwSize msldims[2];
      msldims[0]=ncontrols;
      msldims[1]=1;
      mwSize ndim=2;
      plhs[0] = mxCreateNumericArray(ndim, msldims, mxDOUBLE_CLASS, mxREAL);
      double *mslLengths=mxGetPr(plhs[0]);
      mwSize armdims[2];
      armdims[0]=ncontrols;
      armdims[1]=nU;
      double *mArms;
      if (nlhs>1)
      {
        plhs[1] = mxCreateNumericArray(ndim, armdims, mxDOUBLE_CLASS, mxREAL);
        mArms=mxGetPr(plhs[1]);
      }
      mxArray *Paths;
      if (nlhs>2)
      {
         Paths = mxCreateCellMatrix((mwSize)model->getForceSet().getSize(),3);
         plhs[2] = Paths;
      } 
      else if (nlhs>1)
      {
          Paths = 0;
      }
      else 
      {
         Paths=0;
         mArms=0;
      }

      for (int i=0;i<ncols;++i)
      {
         setState(&stateVec[i*nstates]);
         getLengths(&mslLengths[i*ncontrols],mArms, Paths);
      }
   }
}